import sys
import os

cpu_index = 0
cpu_total = 1
job_index = 0

if len(sys.argv)>=3:
	cpu_index = int(sys.argv[1])
	cpu_total = int(sys.argv[2])
else:
	print("warning : arguments might be missing")

def add_job(func):
	global job_index
	if job_index == cpu_index:
		print(f"I'm running job {job_index}")
		func()
	job_index = (job_index+1)%cpu_total
	# global job_index
	# job_index = job_index+1
	# print(job_index)

###############################################################

from functools import reduce # Valid in Python 2.6+, required in Python 3
import operator

import torch
import torch.nn as nn
import torch.nn.functional as F

import anyact as aa

import numpy as np
np.seterr(divide='ignore', invalid='ignore')

import torchvision
import torchvision.datasets as datasets

import os
import h5py

from mnist import fmnist,mnist,mnist_perf

import matplotlib.pyplot as plt

fmnist_L_x,fmnist_L_y = fmnist(True)
fmnist_T_x,fmnist_T_y = fmnist(False)

def lerp(a,b,λ): return λ*b + (1-λ)*a

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

def nn_fc(layers,σ,dropout=0.5,batchnorm=1):
	layers = [int(np.round(x)) for x in layers]
	disable_drp = 0 if dropout==0 else 1
	disable_btn = batchnorm
	if hasattr(σ,"m"): m=σ.m
	else: m = 1
	return sum([
		[nn.Linear((m if not index==0 else 1)*λ1,λ2),
		]+disable_drp*[aa.Dropout(dropout),
		]+disable_btn*[nn.BatchNorm1d(λ2*m),
		]+[σ,
		]
		for index,(λ1,λ2) in enumerate(zip(layers[0:-2],layers[1:-1]))
	],[])+[nn.Linear(m*layers[-2],layers[-1])] # remove last GELU, BatchNorm and Dropout

# x=torch.randn(5,1,28,28)
# nn.Sequential(nn.Flatten(),*nn_fc([28*28,30,10],aa.Stack([nn.GELU(),nn.GELU()]),dropout=0.5,batchnorm=1))(x).shape


import random
class BatchIterator:
	def __init__(self,tensors,batch_size,shuffle=True):
		self.tensors = tensors # must be a tuple of tensor
		self.batch_size = batch_size
		# sanity check
		if not len(set([t.shape[0] for t in tensors])) == 1: print("ERROR : inconsistent input")
		self.indices = list(range(tensors[0].shape[0]))
		if shuffle: random.shuffle(self.indices)
	def __iter__(self): return self
	def __next__(self):
		batch_size= min(self.batch_size,len(self.indices))
		if batch_size==0: raise StopIteration
		indices = [self.indices.pop(0) for s in range(batch_size)]
		return [torch.stack([t[i] for i in indices],0) for t in self.tensors]

# xs = torch.randn(10,1,28,28)
# ys = torch.randn(10,10)
# for i,(x,y) in enumerate(BatchIterator((xs,ys),3)):
# 	print(x.shape,y.shape)

# def analysis_classification(model,identifier,lr0=-2,lr1=-4,epochs=256,batch_size=1024):
def analysis_classification(model,identifier,loss_fun=nn.MSELoss(),lr0=-2,lr1=-4,epochs=128,batch_size=64):
	model.to(device)
	model.identifier=identifier
	optim = torch.optim.Adam(model.parameters(),lr=10**lr0)
	hist = {
		"epoch_L":[], "loss_L": [], "loss_L_μ": [], "loss_L_σ": [], "perf_L_precision": [], "perf_L_accuracy": [], "perf_L_recall": [], "perf_L_F1": [], "perf_L_cm": [],
		"epoch_T":[], "loss_T": [], "loss_T_μ": [], "loss_T_σ": [], "perf_T_precision": [], "perf_T_accuracy": [], "perf_T_recall": [], "perf_T_F1": [], "perf_T_cm": [],
	}
	for epoch in range(epochs):
		lr = 10**lerp(lr0,lr1,epoch/(epochs-1))
		for g in optim.param_groups: g['lr'] = lr
		#
		model.train()
		losses=[]
		for i,(x,y) in enumerate(BatchIterator((fmnist_L_x,fmnist_L_y),batch_size)):
			x=x.to(device)
			y=y.to(device)
			loss = loss_fun(model(x),y)
			optim.zero_grad();loss.requires_grad_().backward();optim.step()
			losses.append(loss.item())
			hist["loss_L"].append(loss.item())
			hist["epoch_L"].append(epoch + i*batch_size/fmnist_L_x.shape[0])
		hist["loss_L_μ"].append(np.mean(losses))
		hist["loss_L_σ"].append(np.std(losses))
		#
		model.eval()
		losses=[]
		for i,(x,y) in enumerate(BatchIterator((fmnist_T_x,fmnist_T_y),batch_size)):
			x=x.to(device)
			y=y.to(device)
			loss = loss_fun(model(x),y)
			losses.append(loss.item())
			hist["loss_T"].append(loss.item())
			hist["epoch_T"].append(epoch + i*batch_size/fmnist_T_x.shape[0])
		hist["loss_T_μ"].append(np.mean(losses))
		hist["loss_T_σ"].append(np.std(losses))
		#
		rgb_L = (0,.5,1)
		rgb_T = (1,.5,0)
		#
		pr_L,re_L,ac_L,F1_L,cm_L=mnist_perf(model,device,"learn",fmnist_L_x,fmnist_L_y,batch_size,rgb_L) ### on learn set
		hist["perf_L_precision"].append(pr_L)
		hist["perf_L_recall"   ].append(re_L)
		hist["perf_L_accuracy" ].append(ac_L)
		hist["perf_L_F1"       ].append(F1_L)
		hist["perf_L_cm"       ].append(cm_L)
		#
		pr_T,re_T,ac_T,F1_T,cm_T=mnist_perf(model,device,"test" ,fmnist_T_x,fmnist_T_y,batch_size,rgb_T) ### on test set
		hist["perf_T_precision"].append(pr_T)
		hist["perf_T_recall"   ].append(re_T)
		hist["perf_T_accuracy" ].append(ac_T)
		hist["perf_T_F1"       ].append(F1_T)
		hist["perf_T_cm"       ].append(cm_T)
		os.makedirs(f"out/",exist_ok=True)
		print(f"epoch={epoch+1:3d}/{epochs:3d} | log(loss)={np.log(hist['loss_L_μ'][-1]):.5f} | L_prec={pr_L:.5f} L_recall={re_L:.5f} L_accur={ac_L:.5f} L_F1={F1_L:.5f} | T_prec={pr_T:.5f} T_recall={re_T:.5f} T_accur={ac_T:.5f} T_F1={F1_T:.5f}")
		#
		plt.figure()
		plt.scatter(hist["epoch_L"],np.log(hist['loss_L']), label='log(loss) train', s=0.05, alpha=1/4, color=rgb_L)
		plt.scatter(hist["epoch_T"],np.log(hist['loss_T']), label='log(loss) test' , s=0.05, alpha=1/4, color=rgb_T)
		plt.legend()
		plt.tight_layout();plt.savefig(f"out/{model.identifier}-loss.png");plt.close('all')
		#
		plt.figure()
		plt.plot(hist['perf_L_accuracy' ], label=f'learn accuracy' , color=rgb_L , linestyle='-')
		plt.plot(hist['perf_L_precision'], label=f'learn precision', color=rgb_L, linestyle='--')
		plt.plot(hist['perf_L_recall'   ], label=f'learn recall'   , color=rgb_L, linestyle=':') # TODO : why perf_L_recall == perf_L_accuracy ?
		plt.plot(hist['perf_L_F1'       ], label=f'learn F1'       , color=rgb_L, linestyle='-.')
		plt.plot(hist['perf_T_accuracy' ], label=f'test accuracy'  , color=rgb_T, linestyle='-')
		plt.plot(hist['perf_T_precision'], label=f'test precision' , color=rgb_T, linestyle='--')
		plt.plot(hist['perf_T_recall'   ], label=f'test recall'    , color=rgb_T, linestyle=':') # TODO : why perf_T_recall == perf_T_accuracy ?
		plt.plot(hist['perf_T_F1'       ], label=f'test F1'        , color=rgb_T, linestyle='-.')
		plt.legend()
		plt.tight_layout();plt.savefig(f"out/{model.identifier}-perf.png");plt.close('all')
		#
		plt.figure()
		plt.plot(range(epoch+1),hist['loss_L_μ'], label='learn loss', color=rgb_L)
		plt.plot(range(epoch+1),hist['loss_T_μ'], label='test loss' , color=rgb_T)
		plt.fill_between(range(epoch+1), np.array(hist['loss_L_μ'])-np.array(hist['loss_L_σ']), np.array(hist['loss_L_μ'])+np.array(hist['loss_L_σ']), alpha=0.2, color=rgb_L)
		plt.fill_between(range(epoch+1), np.array(hist['loss_T_μ'])-np.array(hist['loss_T_σ']), np.array(hist['loss_T_μ'])+np.array(hist['loss_T_σ']), alpha=0.2, color=rgb_T)
		plt.legend()
		plt.yscale('log')
		plt.tight_layout();plt.savefig(f"out/{model.identifier}-loss-2.png");plt.close('all')
		#
	with h5py.File(f'out/{model.identifier}.h5', 'w') as file:
		for k in hist: file.create_dataset(k, data=hist[k])

def analysis_auto_enc(model,identifier,lr0,lr1,epochs=64,batch_size=64):
	pass # TODO

def expspace(x0,x1,N): return torch.exp(torch.linspace(np.log(x0),np.log(x1),N))

def gradiator(φ,grad_φ):
	class GradiatorFunction(torch.autograd.Function):
		@staticmethod
		def forward(ctx,x):
			y = φ(x)
			ctx.save_for_backward(x,y) # save input for backward pass
			return y
		@staticmethod
		def backward(ctx, grad_o):
			# if not ctx.needs_input_grad[0]: return None # if grad not required, don't compute
			x,y = ctx.saved_tensors # restore input from context
			grad_i = grad_φ(x,y,grad_o)
			return grad_i
	return GradiatorFunction.apply

device = 'cuda'

def cantor():
	if not hasattr(cantor, 'array'): cantor.array = aa.cantor_function(torch.linspace(0,1,3**5+1),5).to(device)
	return aa.modulize(gradiator(
		lambda x: torch.floor(x) + aa.sampler1d_repeat(cantor.array,x,aa.nearest_floor),
		# lambda x: aa.sampler1d_clamp(cantor.array,x,aa.nearest_floor),
		lambda x,y,grad: torch.zeros(x.shape,device=x.device) # derivative is 0 everywhere
	))

def cantor_fakegrad():
	if not hasattr(cantor, 'array'): cantor.array = aa.cantor_function(torch.linspace(0,1,3**5+1),5).to(device)
	return aa.modulize(gradiator(
		lambda x: torch.floor(x) + aa.sampler1d_repeat(cantor.array,x,aa.nearest_floor),
		lambda x,y,grad: grad # fake the gradient to identity
	))

def square_fakegrad():
	return aa.modulize(gradiator(
		lambda x: torch.round(torch.sin(x)),
		lambda x,y,grad: torch.cos(x)*grad # gradient forgets about rounding
	))

class ParametricSin(nn.Module):
	def __init__(self,x_shape=[1]):
		super().__init__()
		self.θ = nn.parameter.Parameter(torch.ones(x_shape))
	def forward(self,x):
		return torch.sin(self.θ*x)

# x = torch.randn(10,4)
# psin = ParametricSin(x.shape)
# psin(x)

for loss_name,loss in [["MSE",nn.MSELoss()],]:
	for σ_name,σ in [["exp_other_lr",aa.modulize(torch.sin)]]:
		for N in [2,4,16]:
			for dropout in [0,0.5]:
				for batchnorm in [1,0]:
					add_job(lambda: analysis_classification(
						nn.Sequential(
							nn.Flatten(),
							*nn_fc(expspace(28*28,10,N+1),σ,dropout=dropout,batchnorm=batchnorm),
							nn.Softmax(dim=1)
						),
						f"fc({N})-{σ_name}-dropout={dropout}-batchnorm={batchnorm}",
						loss,
						lr0=-4,lr1=-6,epochs=256
					))


for loss_name,loss in [["MSE",nn.MSELoss()],]:
	for σ_name,σ in [["sin_other_lr",aa.modulize(torch.sin)]]:
		for N in [2,4,16]:
			for dropout in [0,0.5]:
				for batchnorm in [1,0]:
					add_job(lambda: analysis_classification(
						nn.Sequential(
							nn.Flatten(),
							*nn_fc(expspace(28*28,10,N+1),σ,dropout=dropout,batchnorm=batchnorm),
							nn.Softmax(dim=1)
						),
						f"fc({N})-{σ_name}-dropout={dropout}-batchnorm={batchnorm}",
						loss,
						lr0=-2,lr1=-6,epochs=256
					))

# x = torch.linspace(-3,3,3**10+1)
# x = torch.linspace(-2,2,3**10+1)
# plt.plot(x,cantor_fakegrad()(x.to('cuda')).cpu().numpy())
# plt.plot(x,cantor()(x.to('cuda')).cpu().numpy())
# plt.plot(x,aa.sampler1d_repeat(cantor.array,x.to(device),aa.nearest).cpu().numpy())

############## MAIN STUFF ##############

σs=[
	### https://www.youtube.com/watch?v=Ae9EKCyI1xU
	["Identity",nn.Identity()],
	### RELU LIKE
	["GELU"          ,nn.GELU()],
	["ReLU"          ,nn.ReLU()],
	# ["ReLU6"         ,nn.ReLU6()],
	["RReLU"         ,nn.RReLU(inplace=True)],
	["PReLU"         ,nn.PReLU()],
	["SiLU"          ,nn.SiLU()],
	["ELU"           ,nn.ELU(inplace=True)],
	["SELU"          ,nn.SELU(inplace=True)],
	["GELU"          ,nn.GELU()],
	["CELU"          ,nn.CELU()],
	["Mish"          ,nn.Mish(inplace=True)],
	["Softplus"      ,nn.Softplus()],
	["Hardswish"     ,nn.Hardswish(inplace=True)],
	["LeakyReLU"     ,nn.LeakyReLU(inplace=True)],
	["LeakyReLU(0.1)",nn.LeakyReLU(0.1,inplace=True)],
	["LeakyReLU(0.5)",nn.LeakyReLU(0.5,inplace=True)],
	["LeakyReLU(5.0)",nn.LeakyReLU(5.0,inplace=True)],
	# ["Threshold"  ,nn.Threshold(inplace=True)],
	### CLASSICAL
	["Softsign"   ,nn.Softsign()],
	["Tanh"       ,nn.Tanh()],
	["Hardtanh"   ,nn.Hardtanh(inplace=True)],
	["sigmoid"    ,nn.Sigmoid()],
	["Hardsigmoid",nn.Hardsigmoid(inplace=True)],
	### OTHER - RELU-like but in both direction
	["Softshrink" ,nn.Softshrink()],
	["Hardshrink" ,nn.Hardshrink()],
	["Tanhshrink" ,nn.Tanhshrink()],
	["abs"        ,aa.modulize(torch.abs)],
	### TRIGO
	["atan"       ,aa.modulize(torch.atan)],
	["tan"        ,aa.modulize(torch.tan)],
	["sinc"       ,aa.modulize(torch.sinc)],
	["sin"        ,aa.modulize(torch.sin)],
	["sin(πx)"    ,aa.modulize(lambda x:torch.sin(np.pi*x))],
	### STACKING
	["ExpLog"     ,aa.Stack([aa.Log(),aa.Exp()])], # TODO 
	["IdGELU"     ,aa.Stack([nn.Identity(),nn.GELU()])],
	["IdGELUsin"  ,aa.Stack([nn.Identity(),nn.GELU(),aa.modulize(torch.sin)])],
	["Idsin"      ,aa.Stack([nn.Identity(),aa.modulize(torch.sin)])],
	### PROBABLY INAPPROPRIATE
	["Softmax"    ,nn.Softmax()],
	["Softmin"    ,nn.Softmin()],
	["LogSigmoid" ,nn.LogSigmoid()],
	["LogSoftmax" ,nn.LogSoftmax()],
	### MESSED UP
	["exp"        ,aa.modulize(torch.exp)],
	["log"        ,aa.modulize(torch.log)],
	["log_abs"    ,aa.modulize(lambda x:torch.log(1e-5+torch.abs(x)))],
	["sqrt"       ,aa.modulize(torch.sqrt)],
	["x^2"        ,aa.modulize(lambda x: x*x)],
	# TODO : cantor
	["cantor"     ,cantor()],
	["cantor"     ,cantor_fakegrad()],
	# TODO : perlin noise
	# TODO : GLSL noise function
	# TODO : jeff
	# TODO : aa.Grid(σ1,σ2,σ3...)
	# TODO : aa.Alternate(σ1,σ2,σ3...)
]
# print(len(σs))
# x = torch.randn(20,1,28,28)
# for σ_name,σ in σs:
# 	try:
# 		print("######################")
# 		# print(σ_name,nn_fc(expspace(28*28,10,5+1),σ,dropout=.5,batchnorm=1))
# 		print(
# 			nn.Sequential(
# 				nn.Flatten(),
# 				*nn_fc(expspace(28*28,10,1+1),σ,dropout=.5,batchnorm=0),
# 				nn.Softmax(dim=1)
# 			)(x).shape
# 			,σ_name
# 		)
# 	except:
# 		print("!!!!!!!!!!!!!!!!!!!!!!!", σ_name)

# nn.Sequential(
# 	nn.Flatten(),
# 	*nn_fc(expspace(28*28,10,5+1),aa.Stack([aa.Exp(),aa.Log()]),dropout=.5,batchnorm=0),
# 	nn.Softmax(dim=1)
# )(x).shape


# σs = [
# 	["cantor2"    ,cantor2()],
# 	["cantor"     ,cantor()],
# 	# ["sinc(2x)"   ,modulize(lambda x: torch.sinc(2*x))],
# 	# ["fourier_1",modulize(lambda x: sum([A*torch.sin(2*π*ω*x+φ) for A,ω,φ in [ [0,0,π],[1,1,0],[1,2,0],[1,3,0], ]]))],
# 	# ["fourier_2",modulize(lambda x: sum([A*torch.sin(2*π*ω*x+φ) for A,ω,φ in [ [0,0,π],[1,1,π],[1,2,π/8],[-.2,3,π/4], ]]))],
# ]

# https://humphryscomputing.com/Notes/Neural/chaos.html
	# ["log(exp(-1)+abs(x))"        ,modulize(lambda x: torch.log(math.exp(-1)+torch.abs(x)))],
	# ["log(abs(x))"        ,modulize(lambda x: torch.log(torch.abs(x)))],
	# ["exp(x)"          ,modulize(lambda x: torch.exp(x))],
	# ["exp(x)+log(abs(x))" ,modulize(lambda x: torch.exp(x)+torch.log(torch.abs(x)))],
	# ["exp(sin(x)⁻¹)"  , modulize(lambda x: torch.exp(1/torch.sin(x)))],
	# ["exp(sin(πx)⁻¹)" , modulize(lambda x: torch.exp(1/torch.sin(π*x)))],
	# ["exp(sin(2πx)⁻¹)", modulize(lambda x: torch.exp(1/torch.sin(2*π*x)))],
	# ["exp(sin(3πx)⁻¹)", modulize(lambda x: torch.exp(1/torch.sin(3*π*x)))],
	# ["exp(x)-1"                   ,modulize(lambda x: torch.exp(x)-1)],
	# ["exp(x)-1+log(exp(-1)+abs(x))" ,modulize(lambda x: torch.exp(x)-1+torch.log(math.exp(-1)+torch.abs(x)))],
	# ["-exp(x)+1-log(exp(-1)+abs(x))" ,modulize(lambda x:-torch.exp(x)+1-torch.log(math.exp(-1)+torch.abs(x)))],
	# ["exp(sin(x)⁻¹)-1"  , modulize(lambda x: torch.exp(1/(1+torch.sin(    x))-1))],
	# ["exp(sin(1πx)⁻¹)-1", modulize(lambda x: torch.exp(1/(1+torch.sin(1*π*x))-1))],
	# ["exp(sin(2πx)⁻¹)-1", modulize(lambda x: torch.exp(1/(1+torch.sin(2*π*x))-1))],
	# ["exp(sin(3πx)⁻¹)-1", modulize(lambda x: torch.exp(1/(1+torch.sin(3*π*x))-1))],
	# # # ["sin-mess-a",modulize(lambda x: (2*torch.sin(3/x))+(3*torch.cos(5/x))+(4*torch.sin(6/x))+(1*torch.cos(3/x)))],
	# # # ["sin-mess-b",modulize(lambda x: torch.sin(3/x)*torch.sin(5/(1-x)))],
	# # # ["sin-mess-b",modulize(lambda x: torch.sin(3*x)*torch.sin(5/(1-x)))],
	# # # ["sin-mess-b",modulize(lambda x: torch.sin(2/(1-x)))],
	# # # ["sin-mess-b",modulize(lambda x: torch.sin(2*π/(.5-x**2)))],
	# # # ["sin-mess-b",modulize(lambda x: torch.sin(5/(1-x)))],
	# ["cos(8πx)⋅cos(4πx)",modulize(lambda x: torch.cos(8*π*x)*torch.cos(4*π*x))],
	# ["cos(8πx)⋅cos(4πx)⋅cos(2πx)",modulize(lambda x: torch.cos(8*π*x)*torch.cos(4*π*x)*torch.cos(2*π*x))],
	# ["cos((x-0.0)⁻¹)",modulize(lambda x: torch.cos(1/x))],
	# ["cos((x-0.5)⁻¹)",modulize(lambda x: torch.cos(1/(x-.5)))],
	# ["cos((x-1.0)⁻¹)",modulize(lambda x: torch.cos(1/(x-1)))],
	# ["cos((x-2.0)⁻¹)",modulize(lambda x: torch.cos(1/(x-2)))],
	# ["cos(5(x-2.0)⁻¹)",modulize(lambda x: torch.cos(5/(x-2)))],
	# ["crazy-cos",modulize(lambda x: torch.cos(5/(0.2+1+torch.cos(π*x))))],
	# # ["sin(8πx)⋅sin(4πx)",modulize(lambda x: torch.sin(8*π*x)*torch.sin(4*π*x))],
	# # ["sin(8πx)⋅sin(4πx)⋅sin(2πx)",modulize(lambda x: torch.sin(8*π*x)*torch.sin(4*π*x)*torch.sin(2*π*x))],
	# ["sin((x-0.0)⁻¹)",modulize(lambda x: torch.sin(1/x))],
	# # ["sin((x-0.5)⁻¹)",modulize(lambda x: torch.sin(1/(x-.5)))],
	# ["sin((x-1.0)⁻¹)",modulize(lambda x: torch.sin(1/(x-1)))],
	# # ["sin((x-2.0)⁻¹)",modulize(lambda x: torch.sin(1/(x-2)))],
	# # ["sin(5(x-2.0)⁻¹)",modulize(lambda x: torch.sin(5/(x-2)))],
	# ["cos(πx)⁻¹"  , modulize(lambda x: 1./(torch.cos(π*x)))],
	# ["sin(cos(πx)⁻¹)"  , modulize(lambda x: torch.sin(1./(torch.cos(π*x))))],
	# ["sin(cos(πx⋅2⁻¹)⁻¹)"  , modulize(lambda x: torch.sin(1./(torch.cos(π*x/2))))],

for loss_name,loss in [["MSE",nn.MSELoss()],]:
	for σ_name,σ in σs:
		for N in [2,4,16]:
			for dropout in [0,0.5]:
				for batchnorm in [1,0]:
					add_job(lambda: analysis_classification(
						nn.Sequential(
							nn.Flatten(),
							*nn_fc(expspace(28*28,10,N+1),σ,dropout=dropout,batchnorm=batchnorm),
							nn.Softmax(dim=1)
						),
						f"fc({N})-{σ_name}-dropout={dropout}-batchnorm={batchnorm}",
						loss
					))

# def nn_cn(N,m_size,M,ker_size,σ):
# 	return nn.Sequential(
# 		nn.Conv2d(), nn.BatchNorm2d(), 
# 		nn.Flatten(),
# 		*nn_fc(torch.linspace(28*28,10,N+1),σ,dropout=dropout),
# 		nn.Softmax(dim=1)
# 	),

# def nn_cn(channels,ker_size,σ,m_size=1,dropout=0):
# 	pad1a = (ker_size-1)//2
# 	pad1b = (ker_size  )//2
# 	pad2a = (  m_size-1)//2
# 	pad2b = (  m_size  )//2
# 	channels = [int(c) for c in channels]
# 	return [
# 			nn.Conv2d(λ1,λ2,ker_size)      if index==0 else
# 			Pad((pad1a,pad1b,pad1a,pad1b)) if index==1 else
# 			nn.BatchNorm2d(λ2)             if index==2 else
# 			σ                              if index==3 else
# 			nn.MaxPool2d(m_size,stride=1)  if index==4 else
# 			Pad((pad2a,pad2b,pad2a,pad2b)) if index==5 else
# 			nn.Dropout(dropout)
# 			for λ1,λ2 in zip(channels[:-1],channels[1:])
# 			for index in range(7)
# 		]


# for m_size in range(1,10):
# 	nn.Sequential(*nn_cn([1,64,64],ker_size=3,m_size=m_size,σ=nn.GELU(),dropout=0))(torch.randn(10,1,28,28)).shape




